package it.inaf.oats.ia2.tapschemamanager.businesslayer;

import it.inaf.oats.ia2.tapschemamanager.datalayer.ColumnEntity;
import it.inaf.oats.ia2.tapschemamanager.datalayer.DLUtil;
import it.inaf.oats.ia2.tapschemamanager.datalayer.TableEntity;
import it.inaf.oats.ia2.tapschemamanager.datalayer.TapSchemaHandler;
import java.io.Serializable;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 *
 * @author Sonia Zorba {@literal <zorba at oats.inaf.it>}
 */
public class Table extends EntityWrapper<TableEntity> implements EntityWrapperContainer<Column>, Serializable {

    private static final long serialVersionUID = -9011881651746083862L;

    public static final String UTYPE = "utype";
    public static final String DESCRIPTION = "description";

    private String selectedColumn;
    private final Map<String, Column> columns;

    private final TapSchemaHandler tapSchemaHandler;
    private final String schemaName;
    private final String tableName;

    public Table(TapSchemaHandler tapSchemaHandler, String schemaName, String tableName, TableEntity tableEntity) throws SQLException {
        super(tableEntity, UTYPE, DESCRIPTION);

        addValue(UTYPE, tableEntity.getUtype());
        addValue(DESCRIPTION, tableEntity.getDescription());

        this.tapSchemaHandler = tapSchemaHandler;
        this.schemaName = schemaName;
        this.tableName = tableName;

        // Load all ColumnEntity map
        columns = new TreeMap<>(String.CASE_INSENSITIVE_ORDER);
        List<String> alreadyLoadedColumns = DLUtil.getEntitiesNames(tableEntity.getColumns());
        for (ColumnEntity columnEntity : tapSchemaHandler.getColumnEntities(schemaName, tableName)) {
            Column column = new Column(tapSchemaHandler, columnEntity);
            columns.put(columnEntity.getName(), column);

            if (alreadyLoadedColumns.contains(columnEntity.getName())) {
                column.setHidden(false);
                column.setStatus(Status.ADDED_PERSISTED);
            }
        }

        setStatus(Status.ADDED_PERSISTED);
    }

    public String getSchemaName() {
        return schemaName;
    }

    public String getFullName() {
        return getEntity().getName();
    }

    @Override
    protected void afterSetValue(String key, String value) {
        TableEntity tableEntity = getEntity();
        switch (key) {
            case UTYPE:
                tableEntity.setUtype(value);
                break;
            case DESCRIPTION:
                tableEntity.setDescription(value);
                break;
        }
    }

    @Override
    public List<String> getAddables() {
        List<String> addables = new ArrayList<>();
        for (Column column : columns.values()) {
            if (column.isHidden()) {
                addables.add(column.getName());
            }
        }
        return addables;
    }

    @Override
    public void addEntityWrapper(String name) throws SQLException {
        Column column = columns.get(name);
        if (column.isHidden()) {
            // Add new
            column.setStatus(Status.ADDED_NOT_PERSISTED);
            column.setHidden(false);
        } else {
            // Undo remove
            column.setStatus(Status.ADDED_PERSISTED);
        }
        selectedColumn = name;

        tapSchemaHandler.addColumn(getEntity(), column.getEntity());
    }

    @Override
    public void removeEntityWrapper(String name) {
        String previousColumn = null;
        String followingColumn = null;
        boolean columnFound = false;
        boolean changeSelection = false;

        // loop on visible columns
        for (String columnName : getEntitiesNames()) {
            if (columnFound) {
                followingColumn = columnName;
                break;
            } else {
                if (columnName.equals(name)) {
                    Column column = columns.get(name);
                    if (column.getStatus() == Status.ADDED_NOT_PERSISTED) {
                        column.setHidden(true);
                        changeSelection = true;
                    } else {
                        column.setStatus(Status.TO_REMOVE);
                    }
                    columnFound = true;
                } else {
                    previousColumn = columnName;
                }
            }
        }

        if (changeSelection) {
            if (previousColumn == null && followingColumn == null) {
                selectedColumn = null; // empty column list
            } else {
                selectedColumn = followingColumn == null ? previousColumn : followingColumn;
            }
        }

        tapSchemaHandler.removeColumn(getEntity(), name);
    }

    @Override
    public Column getEntity(String name) {
        return columns.get(name);
    }

    @Override
    public List<String> getEntitiesNames() {
        List<String> columnsNames = new ArrayList<>();
        for (Column column : columns.values()) {
            if (!column.isHidden()) {
                columnsNames.add(column.getName());
            }
        }
        return columnsNames;
    }

    @Override
    protected void setPersisted() {
        super.setPersisted();
        for (Column column : columns.values()) {
            if (column.getStatus() == Status.TO_REMOVE) {
                column.setHidden(true);
            } else {
                column.setPersisted();
            }
        }
    }

    @Override
    public boolean hasChildren() {
        for (Column column : columns.values()) {
            if (!column.isHidden()) {
                return true;
            }
        }
        return false;
    }

    @Override
    public List<Column> getAllEntityWrappers() {
        List<Column> list = new ArrayList<>();
        for (Column column : columns.values()) {
            if (!column.isHidden()) {
                list.add(column);
            }
        }
        return list;
    }

    @Override
    public void selectEntity(String entityName) {
        selectedColumn = entityName;
    }

    @Override
    public String getName() {
        return tableName;
    }

    @Override
    public Column getSelectedEntity() {
        return selectedColumn == null ? null : columns.get(selectedColumn);
    }

    public String getUtype() {
        return getValue(UTYPE);
    }

    public void setUtype(String utype) {
        setValue(UTYPE, utype);
    }

    public String getDescription() {
        return getValue(DESCRIPTION);
    }

    public void setDescription(String description) {
        setValue(DESCRIPTION, description);
    }
}
