package it.inaf.oats.ia2.tapschemamanager.businesslayer;

import it.inaf.oats.ia2.tapschemamanager.datalayer.DBWrapper;
import it.inaf.oats.ia2.tapschemamanager.datalayer.SchemaEntity;
import it.inaf.oats.ia2.tapschemamanager.datalayer.TapSchemaHandler;
import java.io.Serializable;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 *
 * @author Sonia Zorba {@literal <zorba at oats.inaf.it>}
 */
public class TapSchema implements EntityWrapperContainer<Schema>, Serializable {

    private static final long serialVersionUID = 7607955740953928897L;

    private final TapSchemaHandler tapSchemaHandler;

    private final String name;

    private String selectedSchema;
    private final Map<String, Schema> schemas;

    public TapSchema(DBWrapper dbWrapper, String tapSchemaName, boolean exists) throws SQLException {
        this.name = tapSchemaName;
        schemas = new TreeMap<>(String.CASE_INSENSITIVE_ORDER);

        this.tapSchemaHandler = new TapSchemaHandler(dbWrapper, tapSchemaName, exists);

        if (exists) {
            for (SchemaEntity schemaEntity : tapSchemaHandler.getSchemas()) {
                if (!schemaEntity.getName().equals(tapSchemaName)) {
                    Schema schema = new Schema(tapSchemaHandler, schemaEntity);
                    schemas.put(schema.getName(), schema);
                }
            }
            if (!schemas.isEmpty()) {
                selectEntity(schemas.keySet().iterator().next()); // select first schema
            }
        }
    }

    public String getName() {
        return name;
    }

    public void save() throws SQLException {
        tapSchemaHandler.save();

        Iterator<Schema> iterator = schemas.values().iterator();
        while (iterator.hasNext()) {
            Schema schema = iterator.next();
            if (schema.getStatus() == Status.TO_REMOVE) {
                iterator.remove();
            } else {
                schema.setPersisted();
            }
        }
    }

    public Table getSelectedTable() {
        Schema schema = getSelectedEntity();
        return schema == null ? null : schema.getSelectedEntity();
    }

    public Column getSelectedColumn() {
        Table table = getSelectedTable();
        return table == null ? null : table.getSelectedEntity();
    }

    @Override
    public List<String> getAddables() {
        List<String> addables = Util.getAddables(tapSchemaHandler.getAllSchemas(), schemas.keySet());
        if (addables.contains(name)) {
            addables.remove(name);
        }
        return addables;
    }

    @Override
    public void addEntityWrapper(String name) throws SQLException {
        Schema schema = schemas.get(name);
        SchemaEntity schemaEntity;
        if (schema == null) {
            // Add new
            schemaEntity = tapSchemaHandler.getNewSchema(name);
            schema = new Schema(tapSchemaHandler, schemaEntity);
            schemas.put(name, schema);
            schema.setStatus(Status.ADDED_NOT_PERSISTED);
        } else {
            schemaEntity = schema.getEntity();
            // Undo remove
            schema.setStatus(Status.ADDED_PERSISTED);
        }
        selectedSchema = name;

        tapSchemaHandler.addSchema(schemaEntity);
    }

    @Override
    public void removeEntityWrapper(String name) {
        String previousSchema = null;
        boolean changeSelection = false;
        for (Schema schema : schemas.values()) {
            if (schema.getName().equals(name)) {
                if (schema.getStatus() == Status.ADDED_NOT_PERSISTED) {
                    schemas.remove(name);
                    changeSelection = true;
                } else {
                    schema.setStatus(Status.TO_REMOVE);
                }
                break;
            }
            previousSchema = schema.getName();
        }

        if (changeSelection) {
            if (schemas.isEmpty()) {
                selectedSchema = null;
            } else {
                selectedSchema = previousSchema == null ? schemas.keySet().iterator().next() : previousSchema;
            }
        }

        tapSchemaHandler.removeSchema(name);
    }

    @Override
    public Schema getEntity(String name) {
        return schemas.get(name);
    }

    @Override
    public List<String> getEntitiesNames() {
        return Util.setToList(schemas.keySet());
    }

    @Override
    public final void selectEntity(String entityName) {
        selectedSchema = entityName;
        Schema schema = schemas.get(entityName);
        List<String> tables = schema.getEntitiesNames();
        if (!tables.isEmpty()) {
            schema.selectEntity(tables.get(0));
        }
    }

    @Override
    public boolean hasChildren() {
        return !schemas.isEmpty();
    }

    @Override
    public List<Schema> getAllEntityWrappers() {
        return new ArrayList(schemas.values());
    }

    @Override
    public Schema getSelectedEntity() {
        return selectedSchema == null ? null : schemas.get(selectedSchema);
    }

    public TapSchemaHandler getTapSchemaHandler() {
        return tapSchemaHandler;
    }
}
