package it.inaf.oats.ia2.tapschemamanager.datalayer;

import java.io.Serializable;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.Id;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.PrimaryKeyJoinColumn;
import javax.persistence.Table;

/**
 * JPA entity for the table <code>TAP_SCHEMA.keys</code>.
 *
 * @author Sonia Zorba {@literal <zorba at oats.inaf.it>}
 */
@Entity
@Table(name = "\"keys\"")
public class KeyEntity implements Serializable {

    private static final long serialVersionUID = 4321901642780866919L;

    @Id
    @Column(name = "key_id", length = 64)
    private String keyId;

    @Column(name = "from_table", length = 128)
    private String fromTableFullName;

    @Column(name = "target_table", length = 128)
    private String targetTableFullName;

    @Column(name = "utype", length = 512)
    private String utype;

    @Column(name = "description", length = 512)
    private String description;

    @Column(name = "keyID")
    private BigInteger keyID;

    @ManyToOne(fetch = FetchType.EAGER)
    @PrimaryKeyJoinColumn(name = "from_table", referencedColumnName = "table_name")
    private TableEntity fromTable;

    @ManyToOne(fetch = FetchType.EAGER)
    @PrimaryKeyJoinColumn(name = "target_table", referencedColumnName = "table_name")
    private TableEntity targetTable;

    @OneToMany(mappedBy = "key", cascade = CascadeType.ALL, orphanRemoval = true)
    private final List<KeyColumnEntity> keyColumns;

    private String fromSchemaName;
    private String targetSchemaName;

    public KeyEntity() {
        keyColumns = new ArrayList<>();
    }

    public KeyEntity(String fromSchema, String fromTable, String targetSchema, String targetTable) {
        this();
        this.fromSchemaName = fromSchema;
        this.fromTableFullName = fromSchema + "." + fromTable;
        this.targetSchemaName = targetSchema;
        this.targetTableFullName = targetSchema + "." + targetTable;
    }

    /**
     * The TAP standard define the key ID as a VARCHAR, but in our
     * implementation we assume it represents a numeric value
     *
     * @param keyId - ID of the key
     */
    public KeyEntity(int keyId) {
        this();
        this.keyId = keyId + "";
    }

    /**
     * The value in the <code>key_id</code> column: represents the ID of the key
     * (in our implementation we assume it represents a numeric value).
     */
    public String getKeyId() {
        return keyId;
    }

    public void setKeyId(String keyId) {
        this.keyId = keyId;
    }

    /**
     * The value in the <code>from_table</code> column: it is the complete name
     * of the table, as defined in the TAP standard: so it is
     * <code>&lt;schema-name&gt;.&lt;table-name&gt;</code>.
     */
    public String getFromTableFullName() {
        return fromTableFullName;
    }

    public void setFromTableFullName(String fromTableFullName) {
        this.fromTableFullName = fromTableFullName;
    }

    /**
     * The value in the <code>target_table</code> column: it is the complete
     * name of the table, as defined in the TAP standard: so it is
     * <code>&lt;schema-name&gt;.&lt;table-name&gt;</code>.
     */
    public String getTargetTableFullName() {
        return targetTableFullName;
    }

    public void setTargetTableFullName(String targetTableFullName) {
        this.targetTableFullName = targetTableFullName;
    }

    /**
     * The value in the <code>utype</code> column.
     */
    public String getUtype() {
        return utype;
    }

    public void setUtype(String utype) {
        this.utype = utype;
    }

    /**
     * The value in the <code>description</code> column.
     */
    public String getDescription() {
        return description;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    /**
     * The value in the <code>keyID</code> column: it represents [TODO]...
     */
    public BigInteger getKeyID() {
        return keyID;
    }

    public void setKeyID(BigInteger keyID) {
        this.keyID = keyID;
    }

    /**
     * @return the {@link TableEntity} related to the <code>from_table</code>
     * field
     */
    public TableEntity getFromTable() {
        return fromTable;
    }

    public void setFromTable(TableEntity fromTable) {
        this.fromTable = fromTable;
    }

    /**
     * @return the {@link TableEntity} related to the <code>target_table</code>
     * field
     */
    public TableEntity getTargetTable() {
        return targetTable;
    }

    public void setTargetTable(TableEntity targetTable) {
        this.targetTable = targetTable;
    }

    /**
     * List of the <code>KeyColumnEntity</code> owned by this
     * <code>KeyEntity</code>.
     */
    public List<KeyColumnEntity> getKeyColumns() {
        return keyColumns;
    }

    public String getFromSchemaName() {
        return fromSchemaName;
    }

    public void setFromSchemaName(String fromSchemaName) {
        this.fromSchemaName = fromSchemaName;
    }

    public String getTargetSchemaName() {
        return targetSchemaName;
    }

    public void setTargetSchemaName(String targetSchemaName) {
        this.targetSchemaName = targetSchemaName;
    }
}
