package it.inaf.oats.ia2.tapschemamanager.datalayer;

import java.io.Serializable;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.regex.Pattern;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.Id;
import javax.persistence.ManyToOne;
import javax.persistence.MapKeyColumn;
import javax.persistence.OneToMany;
import javax.persistence.PrimaryKeyJoinColumn;
import javax.persistence.Table;

/**
 * JPA entity for the table <code>TAP_SCHEMA.tables</code>.
 *
 * @author Sonia Zorba {@literal <zorba at oats.inaf.it>}
 */
@Entity
@Table(name = "tables")
public class TableEntity implements Serializable, TapSchemaEntity {

    private static final long serialVersionUID = 5386350231967268073L;

    @Id
    @Column(name = "table_name", length = 128)
    private String name;

    @Column(name = "schema_name", length = 64)
    private String schemaName;

    @Column(name = "table_type", length = 8)
    private String tableType;

    @Column(name = "utype", length = 512)
    private String utype;

    @Column(name = "description", length = 512)
    private String description;

    @Column(name = "tableID")
    private BigInteger tableID;

    @ManyToOne(fetch = FetchType.EAGER)
    @PrimaryKeyJoinColumn(name = "schema_name", referencedColumnName = "schema_name")
    private SchemaEntity schema;

    @OneToMany(mappedBy = "table", fetch = FetchType.EAGER, cascade = CascadeType.ALL, orphanRemoval = true)
    @MapKeyColumn(name = "column_name", insertable = false, updatable = false)
    private final List<ColumnEntity> columns;

    @OneToMany(mappedBy = "fromTable", fetch = FetchType.EAGER, cascade = CascadeType.ALL, orphanRemoval = true)
    private final List<KeyEntity> fromKeys;

    @OneToMany(mappedBy = "targetTable", fetch = FetchType.EAGER, cascade = CascadeType.ALL, orphanRemoval = true)
    private final List<KeyEntity> targetKeys;

    protected TableEntity() {
        columns = new ArrayList<>();
        fromKeys = new ArrayList<>();
        targetKeys = new ArrayList<>();
    }

    protected TableEntity(SchemaEntity schemaEntity, String shortTableName) {
        this();
        this.schema = schemaEntity;
        this.schemaName = schemaEntity.getName();
        this.name = schemaName + "." + shortTableName;
    }

    /**
     * The value in the <code>table_name</code> column: it represents the
     * complete table name: <code>schema_name.table_name</code>.
     */
    @Override
    public String getName() {
        return name;
    }

    /**
     * The table name without the schema name.
     */
    public String getShortTableName() {
        return name.split(Pattern.quote("."))[1];
    }

    protected void setName(String fullTableName) {
        this.name = fullTableName;
    }

    /**
     * The value in the <code>schema_name</code> column.
     */
    public String getSchemaName() {
        return schemaName;
    }

    protected void setSchemaName(String schemaName) {
        this.schemaName = schemaName;
    }

    /**
     * The value in the <code>table_type</code> column: it could be
     * <code>table</code> or <code>view</code>.
     */
    public String getTableType() {
        return tableType;
    }

    public void setTableType(String tableType) {
        this.tableType = tableType;
    }

    /**
     * The value in the <code>utype</code> column.
     */
    public String getUtype() {
        return utype;
    }

    public void setUtype(String utype) {
        this.utype = utype;
    }

    /**
     * The value in the <code>description</code> column.
     */
    public String getDescription() {
        return description;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    /**
     * The value in the <code>tableID</code> column: it represents [TODO]...
     */
    protected BigInteger getTableID() {
        return tableID;
    }

    protected void setTableID(BigInteger tableID) {
        this.tableID = tableID;
    }

    /**
     * The {@link SchemaEntity} that owns this TableEntity.
     */
    public SchemaEntity getSchema() {
        return schema;
    }

    protected void setSchema(SchemaEntity schemaEntity) {
        this.schema = schemaEntity;
    }

    /**
     * The list of <code>ColumnEntity</code> owned by this
     * <code>TableEntity</code>.
     *
     * @return an unmodifiable list
     */
    public List<ColumnEntity> getColumns() {
        return Collections.unmodifiableList(columns);
    }

    /**
     * The list of <code>KeyEntity</code> which <code>from_table</code> column
     * is the complete name of this <code>TableEntity</code>.
     *
     * @return an unmodifiable list
     */
    public List<KeyEntity> getFromKeys() {
        return Collections.unmodifiableList(fromKeys);
    }

    /**
     * The list of <code>KeyEntity</code> which <code>target_table</code> column
     * is the complete name of this <code>TableEntity</code>.
     *
     * @return an unmodifiable list
     */
    public List<KeyEntity> getTargetKeys() {
        return Collections.unmodifiableList(targetKeys);
    }

    protected void addColumn(ColumnEntity column) {
        columns.add(column);
        column.setFullTableName(name);
        column.setTable(this);
    }

    protected void removeColumn(String columnName) {
        DLUtil.removeEntity(columns, columnName);
    }

//    public boolean containsKey(KeyEntity keyEntity) {
//        String keyId = keyEntity.getKeyId();
//        for (KeyEntity key : fromKeys) {
//            if (key.getKeyId().equals(keyId)) {
//                return true;
//            }
//        }
//        for (KeyEntity key : targetKeys) {
//            if (key.getKeyId().equals(keyId)) {
//                return true;
//            }
//        }
//        return false;
//    }
    public void addFromKey(KeyEntity key) {
        if (!fromKeys.contains(key)) {
            fromKeys.add(key);
        }
    }

    public void removeFromKey(KeyEntity key) {
        fromKeys.remove(key);
    }

    public void addTargetKey(KeyEntity key) {
        if (!targetKeys.contains(key)) {
            targetKeys.add(key);
        }
    }

    public void removeTargetKey(KeyEntity key) {
        targetKeys.remove(key);
    }
}
