package it.inaf.oats.ia2.tapschemamanager.datalayer;

import java.io.Closeable;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.sql.DataSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

class SchemaInfo implements Serializable {

    private static final long serialVersionUID = 2835388936102349727L;

    private final List<String> allSchemaTables;
    private final List<KeyEntity> schemaKeys;
    private final Map<String, String> tablesType;

    SchemaInfo(DataSource dataSource, DatabaseType dbType, String schemaName) throws SQLException {
        allSchemaTables = Dao.getAllTablesNames(dataSource, dbType, schemaName);
        schemaKeys = Dao.getSchemaKeys(dataSource, dbType, schemaName);
        tablesType = Dao.getTablesTypes(dataSource, dbType, schemaName);
    }

    public List<String> getAllSchemaTables() {
        return allSchemaTables;
    }

    public List<KeyEntity> getSchemaKeys() {
        return schemaKeys;
    }

    public Map<String, String> getTablesType() {
        return tablesType;
    }
}

class ColumnsInfo implements Serializable {

    private static final long serialVersionUID = -3407052226160044192L;

    private final String tapSchemaName;
    private final Map<String, Map<String, List<ColumnEntity>>> columnsInfo;

    ColumnsInfo(String tapSchemaName) {
        this.tapSchemaName = tapSchemaName;
        this.columnsInfo = new HashMap<>();
    }

    void addColumnsInfo(DBWrapper dbWrapper, TableEntity tableEntity) throws SQLException {
        String schemaName = tableEntity.getSchemaName();
        Map<String, List<ColumnEntity>> map = columnsInfo.get(schemaName);
        if (map == null) {
            map = new HashMap<>();
            columnsInfo.put(schemaName, map);
        }

        List<ColumnEntity> allColumns;
        if (schemaName.equals(tapSchemaName)) {
            allColumns = Dao.getAllColumns(dbWrapper.getTapSchemaDataSource(), dbWrapper.getTapSchemaDatabaseType(), tableEntity);
        } else {
            allColumns = Dao.getAllColumns(dbWrapper.getSourceDataSource(), dbWrapper.getSourceDatabaseType(), tableEntity);
        }
        map.put(tableEntity.getShortTableName(), allColumns);
    }

    List<ColumnEntity> getColumnsInfo(String schemaName, String shortTableName) {
        return columnsInfo.get(schemaName).get(shortTableName);
    }
}

/**
 * This class manage load of TAP_SCHEMA entities and other database schemas
 * informations. IMPORTANT: If you use this class inside another object remember
 * to call the close method to release the occupied resources.
 *
 * @author Sonia Zorba {@literal <zorba at oats.inaf.it>}
 */
public class TapSchemaHandler implements Serializable, Closeable {

    private static final long serialVersionUID = 1193743698810532785L;
    private static final Logger log = LoggerFactory.getLogger(TapSchemaHandler.class);

    private final DBWrapper dbWrapper;

    //private final Credentials sourceCredentials;
    //private final Credentials tapSchemaCredentials;
    private final String tapSchemaName;
    private boolean exists;

    private final List<String> allSchemas;
    private final List<SchemaEntity> schemas;
    private final List<String> persistedSchemas;
    private final List<SchemaEntity> toRemoveSchemas;

    private final Map<String, SchemaInfo> allSchemaInfos;

    private final ColumnsInfo columnsInfo;

    private transient EntityManager entityManager;

    public TapSchemaHandler(DBWrapper dbWrapper, String tapSchemaName, boolean exists) throws SQLException {
        this.dbWrapper = dbWrapper;
        this.tapSchemaName = tapSchemaName;
        this.exists = exists;

        this.allSchemas = Dao.getAllSchemasNames(dbWrapper.getSourceDataSource(), dbWrapper.getSourceDatabaseType());

        this.schemas = new ArrayList<>();
        this.persistedSchemas = new ArrayList<>();
        this.toRemoveSchemas = new ArrayList<>();

        this.allSchemaInfos = new HashMap<>();

        this.columnsInfo = new ColumnsInfo(tapSchemaName);

        if (exists) {
            loadEntityManager();
            CriteriaBuilder cb = entityManager.getCriteriaBuilder();
            CriteriaQuery<SchemaEntity> cq = cb.createQuery(SchemaEntity.class);
            // Load all SchemaEntities saved into the TAP_SCHEMA
            for (SchemaEntity schemaEntity : entityManager.createQuery(cq).getResultList()) {
                String schemaName = schemaEntity.getName();
                addSchema(schemaEntity);

                SchemaInfo schemaInfo;
                if (schemaName.equals(tapSchemaName)) {
                    schemaInfo = new SchemaInfo(dbWrapper.getTapSchemaDataSource(), dbWrapper.getTapSchemaDatabaseType(), schemaName);
                } else {
                    schemaInfo = new SchemaInfo(dbWrapper.getSourceDataSource(), dbWrapper.getSourceDatabaseType(), schemaName);
                }

                allSchemaInfos.put(schemaName, schemaInfo);

                persistedSchemas.add(schemaName);
                for (TableEntity tableEntity : schemaEntity.getTables()) {
                    columnsInfo.addColumnsInfo(dbWrapper, tableEntity);
                }
            }
        }
    }

    public SchemaEntity getNewSchema(String schemaName) throws SQLException {
        SchemaEntity schemaEntity = new SchemaEntity(schemaName);
        allSchemaInfos.put(schemaName, new SchemaInfo(dbWrapper.getSourceDataSource(), dbWrapper.getSourceDatabaseType(), schemaName));
        return schemaEntity;
    }

    public final void addSchema(SchemaEntity schemaEntity) {
        schemas.add(schemaEntity);

        for (TableEntity tableEntity : schemaEntity.getTables()) {
            addTableKeys(schemaEntity, tableEntity);
        }
    }

    public final void removeSchema(String schemaName) {
        for (TableEntity tableEntity : DLUtil.getEntityByName(schemas, schemaName).getTables()) {
            removeKeys(tableEntity);
        }

        if (persistedSchemas.contains(schemaName)) {
            toRemoveSchemas.add(getSchema(schemaName));
            persistedSchemas.remove(schemaName);
        }
        DLUtil.removeEntity(schemas, schemaName);
    }

    public List<SchemaEntity> getSchemas() {
        return schemas;
    }

    public List<String> getAllSchemas() {
        return allSchemas;
    }

    public List<String> getAllTables(String schemaName) {
        return allSchemaInfos.get(schemaName).getAllSchemaTables();
    }

    public List<ColumnEntity> getColumnEntities(String schemaName, String shortTableName) {
        return columnsInfo.getColumnsInfo(schemaName, shortTableName);
    }

    public int getMaxKey() {
        int maxKey = 0;
        for (SchemaEntity schemaEntity : schemas) {
            for (TableEntity tableEntity : schemaEntity.getTables()) {
                for (KeyEntity keyEntity : tableEntity.getFromKeys()) {
                    int key = Integer.parseInt(keyEntity.getKeyId());
                    if (key > maxKey) {
                        maxKey = key;
                    }
                }
                for (KeyEntity keyEntity : tableEntity.getTargetKeys()) {
                    int key = Integer.parseInt(keyEntity.getKeyId());
                    if (key > maxKey) {
                        maxKey = key;
                    }
                }
            }
        }
        return maxKey;
    }

    public TableEntity getNewTable(String schemaName, String shortTableName) throws SQLException {
        TableEntity table = new TableEntity(getSchema(schemaName), shortTableName);

        table.setTableType(allSchemaInfos.get(schemaName).getTablesType().get(shortTableName));

        // Add columns info
        // For performance reasons all possible column entities are loaded together
        columnsInfo.addColumnsInfo(dbWrapper, table);

        return table;
    }

    private void addTableKeys(SchemaEntity schemaEntity, TableEntity tableEntity) {
        int currentKey = getMaxKey() + 1;
        for (KeyEntity keyEntity : getAllSchemaKeys(schemaEntity.getName())) {

            String fromTableSchemaName = keyEntity.getFromSchemaName();
            String targetTableSchemaName = keyEntity.getTargetSchemaName();

            boolean fromTableSchemaExposed = false;
            boolean targetTableSchemaExposed = false;
            for (SchemaEntity s : getSchemas()) {
                if (fromTableSchemaExposed && targetTableSchemaExposed) {
                    break;
                }
                if (s.getName().equals(fromTableSchemaName)) {
                    fromTableSchemaExposed = true;
                }
                if (s.getName().equals(targetTableSchemaName)) {
                    targetTableSchemaExposed = true;
                }
            }

            // If both schemas are exposed it is possible to add the key
            if (fromTableSchemaExposed && targetTableSchemaExposed) {
                String fromTableName = keyEntity.getFromTableFullName();
                String targetTableName = keyEntity.getTargetTableFullName();
                TableEntity fromTable = null, targetTable = null;
                if (tableEntity.getName().equals(fromTableName)) {
                    fromTable = tableEntity;
                    targetTable = schemaEntity.getTable(targetTableName);
                } else if (tableEntity.getName().equals(targetTableName)) {
                    targetTable = tableEntity;
                    fromTable = schemaEntity.getTable(fromTableName);
                }
                if (fromTable != null && targetTable != null
                        && !fromTable.getFromKeys().contains(keyEntity)
                        && !targetTable.getTargetKeys().contains(keyEntity)) {
                    String key = currentKey + "";
                    keyEntity.setKeyId(key);
                    keyEntity.setFromTable(fromTable);
                    keyEntity.setTargetTable(targetTable);
                    for (KeyColumnEntity keyColumn : keyEntity.getKeyColumns()) {
                        keyColumn.setKeyId(key);
                    }

                    fromTable.addFromKey(keyEntity);
                    targetTable.addTargetKey(keyEntity);

                    currentKey++;
                }
            }
        }
    }

    public void addTable(String schemaName, TableEntity tableEntity) {
        SchemaEntity schemaEntity = getSchema(schemaName);
        schemaEntity.addTable(tableEntity);
        addTableKeys(schemaEntity, tableEntity);
    }

    private void removeKeys(TableEntity tableEntity) {
        for (KeyEntity keyEntity : tableEntity.getFromKeys()) {
            keyEntity.getTargetTable().removeTargetKey(keyEntity);
        }
        for (KeyEntity keyEntity : tableEntity.getTargetKeys()) {
            keyEntity.getFromTable().removeFromKey(keyEntity);
        }
    }

    public void removeTable(SchemaEntity schemaEntity, TableEntity tableEntity) {
        removeKeys(tableEntity);
        schemaEntity.removeTable(tableEntity.getName());
    }

    public void addColumn(TableEntity tableEntity, ColumnEntity columnEntity) {
        tableEntity.addColumn(columnEntity);
    }

    public void removeColumn(TableEntity tableEntity, String columnName) {
        tableEntity.removeColumn(columnName);
    }

    public String getForeignKeyReference(String schemaName, String fullTableName, String columnName) {
        for (KeyEntity key : getAllSchemaKeys(schemaName)) {
            if (key.getFromTableFullName().equals(fullTableName)) {
                for (KeyColumnEntity keyColumn : key.getKeyColumns()) {
                    if (keyColumn.getFromColumn().equals(columnName)) {
                        return key.getTargetTableFullName().split(Pattern.quote("."))[1] + "." + keyColumn.getTargetColumn();
                    }
                }
            }
        }

        return null;
    }

    public SchemaEntity getSchema(String schemaName) {
        return DLUtil.getEntityByName(schemas, schemaName);
    }

    public List<KeyEntity> getAllSchemaKeys(String schemaName) {
        return allSchemaInfos.get(schemaName).getSchemaKeys();
    }

    public void save() throws SQLException {
        if (exists) {
            log.info("Saving existing TAP_SCHEMA {}", tapSchemaName);

            loadEntityManager();

            entityManager.getTransaction().begin();
            for (SchemaEntity toRemoveSchema : toRemoveSchemas) {
                entityManager.remove(toRemoveSchema);
            }
            for (SchemaEntity schemaEntity : schemas) {
                entityManager.merge(schemaEntity);
            }
            entityManager.getTransaction().commit();
        } else {
            log.info("Saving TAP_SCHEMA {} for the first time", tapSchemaName);
            Dao.createTapSchemaSchema(dbWrapper, tapSchemaName);

            loadEntityManager();

            entityManager.getTransaction().begin();
            for (SchemaEntity schemaEntity : schemas) {
                entityManager.persist(schemaEntity);
            }
            entityManager.getTransaction().commit();

            // Adding TAP_SCHEMA into TAP_SCHEMA
            log.info("Adding TAP_SCHEMA into TAP_SCHEMA");
            SchemaEntity tapSchemaSchema;
            tapSchemaSchema = getNewSchema(tapSchemaName);
            addSchema(tapSchemaSchema);

            for (String shortTableName : allSchemaInfos.get(tapSchemaName).getAllSchemaTables()) {
                TableEntity tableEntity = getNewTable(tapSchemaName, shortTableName);
                addTable(tapSchemaName, tableEntity);
                for (ColumnEntity ce : columnsInfo.getColumnsInfo(tapSchemaName, shortTableName)) {
                    addColumn(tableEntity, ce);
                }
            }

            DLUtil.putInfoIntoTapSchemaSchema(tapSchemaSchema);

            entityManager.getTransaction().begin();
            entityManager.merge(tapSchemaSchema);
            entityManager.getTransaction().commit();

            exists = true;
        }
    }

    private void loadEntityManager() {
        if (entityManager == null) {
            entityManager = Dao.getEntityManager(dbWrapper.getTapSchemaCredentials(), tapSchemaName);
        }
    }

    @Override
    public void close() {

        if (entityManager != null && entityManager.isOpen()) {
            EntityManagerFactory efFactory = entityManager.getEntityManagerFactory();

            entityManager.close();

            if (efFactory.isOpen()) {
                efFactory.close();
            }
        }
    }

    /**
     * Automatic closing on serialization
     *
     * @param stream
     * @throws IOException
     */
    private void writeObject(ObjectOutputStream stream) throws IOException {
        close();
        stream.defaultWriteObject();
    }
}
